Components.utils.import("resource://komodosync/engines/apicatalogs.js");
Components.utils.import("resource://komodosync/engines/keybindings.js");
Components.utils.import("resource://komodosync/engines/kopasswords.js");
Components.utils.import("resource://komodosync/engines/koprefs.js");
Components.utils.import("resource://komodosync/engines/projecttemplates.js");
Components.utils.import("resource://komodosync/engines/schemes.js");
Components.utils.import("resource://komodosync/engines/templates.js");
Components.utils.import("resource://komodosync/engines/tools.js");

Components.utils.import("resource://services-sync/engines.js");
Components.utils.import("resource://services-sync/ext/Observers.js");

var KomodoSync = {
    init: function KomodoSync_init() {
        if (this._migrateTo0_4()) {
            Weave.Service.onStartup();
            return;
        }
        var engines = [
            new ApiCatalogsEngine(),
            new KeybindingsEngine(),
            new KoPasswordEngine("kopasswords"),
            new KoPrefsEngine(),
            new ProjectTemplatesEngine(),
            new SchemesEngine(),
            new TemplatesEngine(),
            new ToolsEngine(),
        ];

        // Monitor sync setup state to show/hide the status bar button (bug 100067)
        // We observing weave:service:ready is better, but that fires before we
        // get constructed on subsequent startups.  So use login attempt as a proxy.
        Weave.Svc.Obs.add("weave:service:login:start", this.onLoginStart, this);
        Weave.Svc.Obs.add("weave:service:start-over", this.onStartOver, this);
        
        Engines.register(engines);
        Weave.Svc.Obs.notify("weave:engine:start-tracking"); // weave triggers this before service:ready
    },
    
    openPrefs: function KomodoSync_openPrefs() {
        let win = Weave.Svc.WinMediator.getMostRecentWindow("KomodoSync:Prefs");
        if (win)
          win.focus();
        else {
            ko.windowManager.openDialog(
                "chrome://weave/content/komodo/prefs-dialog.xul",
                "KomodoSync:Prefs",
                "chrome,titlebar,resizable,close,centerscreen,dialog=yes"
            );
        }
    },
    
    isPrimaryWindow: function KomodoSync_isPrimaryWindow() {
        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
                       .getService(Components.interfaces.nsIWindowMediator);
        return window == wm.getMostRecentWindow("Komodo");
    },
    
    _migrateTo0_4: function KomodoSync__migrateTo0_4() {
        var migrated = false;
        var migrated0_4 = "komodo.migrated.0.4";
        if (!Weave.Svc.Prefs.get(migrated0_4, false)) {
            Weave.Service.clusterURL = "";
            Weave.Svc.Prefs.reset('serverURL');
            Weave.Svc.Prefs.set(migrated0_4, true);
            migrated = true;
        }
        
        return migrated;
    },

    onLoginStart: function KomodoSync_onLoginStart() {
        // Sync has loaded, show the sync notification button (bug 100067)
        document.getElementById("sync-status-button").collapsed = false;
    },
    onStartOver: function KomodoSync_onStartOver() {
        // User has removed the sync login, hide the sync notification button
        // (bug 100067)
        document.getElementById("sync-status-button").collapsed = true;
    },
};

// This is a magic variable name for some of the patches in weave-ext:
var WeaveApp = KomodoSync;

Observers.add("weave:service:ready", KomodoSync.init, KomodoSync);