/* Copyright (c) 2000-2013 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* Komodo's Refactoring/Rename-Attribute first dialog
 * Based on the Find and Replace dialog (rev 2).
 */

//---- globals

var { classes: Cc, interfaces: Ci, utils: Cu } = Components;

var log = ko.logging.getLogger("renameClassMember.dialog");
log.setLevel(ko.logging.LOG_DEBUG);

var widgets = null; // object storing interesting XUL element references
var gFindSvc = null;
var _g_prefs = null;

var _g_searchText = null;
var _g_codeintelDefn = null;
var _g_view = null;
var _g_refactoringLanguageObj = null;
 
var _bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
      .getService(Components.interfaces.nsIStringBundleService)
      .createBundle("chrome://komodo/locale/find/find2.properties");

const koIFindOptions = Ci.koIFindOptions;

//---- public methods for the dialog

function onLoad() {
    try {
        _g_prefs = Components.classes["@activestate.com/koPrefService;1"]
            .getService(Components.interfaces.koIPrefService).prefs;
        gFindSvc = Components.classes["@activestate.com/koFindService;1"].
                   getService(Components.interfaces.koIFindService);
        _init_widgets();
        // Necessary for re-launching (i.e. Ctrl+F when the dialog is already open).
        window.focus();
        _init();
        update(null);
    } catch (ex) {
        log.exception(ex);
    }
}

/**
 * Update as appropriate for some change in the dialog.
 *
 * @param {string} changed The name of the thing that changed. If
 *      null or not specified *everything* is updated (used for dialog
 *      initialization).
 */
function update(changed /* =null */) {
    if (typeof(changed) == "undefined") changed = null;
    
    var ui_changed = false;
    var opts = gFindSvc.options;
    
    if (changed == "repl" || changed === null) {
        widgets.replace_all_btn.disabled = !widgets.repl.value;
    }

    if (changed == "case") {
        // Skip this for initialization (changed=null).

        // Advance the checkbox to the next state.
        switch (widgets.opt_case.value) {
        case "ignore-case":
            _set_case_widget("match-case");
            break;
        case "match-case":
            _set_case_widget("smart-case");
            break;
        case "smart-case":
            _set_case_widget("ignore-case");
            break;
        }

        // Save the current state on the global find options.
        switch (widgets.opt_case.value) {
        case "ignore-case":
            opts.caseSensitivity = koIFindOptions.FOC_INSENSITIVE;
            break;
        case "match-case":
            opts.caseSensitivity = koIFindOptions.FOC_SENSITIVE;
            break;
        case "smart-case":
            opts.caseSensitivity = koIFindOptions.FOC_SMART;
            break;
        }
    }
    if (changed == null || changed == "dirs") {
        opts.encodedFolders = widgets.dirs.value;
    }
    if (changed == null || changed == "search-in-subdirs") {
        opts.searchInSubfolders = widgets.search_in_subdirs.checked;
    }
    if (changed == null || changed == "includes") {
        opts.encodedIncludeFiletypes = widgets.includes.value;
    }
    if (changed == null || changed == "excludes") {
        opts.encodedExcludeFiletypes = widgets.excludes.value;
    }
}

/**
 * Change the "Search in:" menulist to the given value.
 */
function search_in(value)
{
    try {
        widgets.search_in_menu.value = value;
        update('search-in');
    } catch(ex) {
        log.exception(ex);
    }
}


/**
 * Handle the onfocus event on the 'dirs' textbox.
 */
function dirs_on_focus(widget, event)
{
    try {
        widget.setSelectionRange(0, widget.textLength);
        // For textbox-autocomplete (TAC) of directories on this widget we
        // need a cwd with which to interpret relative paths. The chosen
        // cwd is that of the current file in the main editor window.
        if (event.target.nodeName == 'html:input') { 
            var textbox = widget.parentNode.parentNode.parentNode;
            var cwd = ko.windowManager.getMainWindow().ko.window.getCwd();
            textbox.searchParam = ko.stringutils.updateSubAttr(
                textbox.searchParam, 'cwd', cwd);            
        }
    } catch(ex) {
        log.exception(ex);
    }
}


function browse_for_dirs() {
    try {
        dump("_g_codeintelDefn.path: " + _g_codeintelDefn.path + "\n");
        var currDir = opener.ko.uriparse.dirName(_g_codeintelDefn.path);
        dump("Use currDir: " + currDir + "\n");
        var obj = {
            encodedFolders: widgets.dirs.value,
            cwd: currDir};
        window.openDialog("chrome://komodo/content/find/browseForDirs.xul",
                          "_blank",
                          "chrome,modal,titlebar,resizable",
                          obj);
        if (obj.retval != "Cancel") {
            widgets.dirs.value = obj.encodedFolders;
            update("dirs");
        }
    } catch(ex) {
        log.exception(ex);
    }
}

function replace_all() {
    try {
        var pattern = widgets.curr_pattern.value;
        if (! pattern) {
            return;
        }
        var repl = widgets.curr_repl.value;
        const addFromACTextbox = false;

        if (addFromACTextbox) {
        ko.mru.addFromACTextbox(widgets.pattern);
        ko.mru.addFromACTextbox(widgets.repl);
        }

        // Always reset the find session for replace all
        var findSessionSvc = Components.classes["@activestate.com/koFindSession;1"].
            getService(Components.interfaces.koIFindSession);
        findSessionSvc.Reset();

        if (addFromACTextbox) {
        ko.mru.addFromACTextbox(widgets.dirs);
        if (widgets.includes.value)
            ko.mru.addFromACTextbox(widgets.includes);
        if (widgets.excludes.value)
            ko.mru.addFromACTextbox(widgets.excludes);
        }
        ko.windowManager.openDialog(
        //XXX: Rename the XUL file
            "chrome://refactoring/content/confirmRenameAttributeChanges.xul",
                    "komodo_refactoring_confirmRenameAttributeChanges",
                    "chrome,close=yes,centerscreen,modal,resizable",
                    {
                            pattern: _g_searchText,
                            repl: widgets.repl.value,
                            matchCase: widgets.opt_case.checked,
                            dirs: widgets.dirs.value,
                            search_in_subdirs: widgets.search_in_subdirs.checked,
                            includes: widgets.includes.value,
                            excludes: widgets.excludes.value,
                            defn: _g_codeintelDefn,
                            view: _g_view,
                            ko: opener.ko,
                            refactoringLanguageObj: _g_refactoringLanguageObj
                            
                    });
        window.close();
    } catch (ex) {
        log.exception(ex);
    }
}



//---- internal support stuff

// Load the global 'widgets' object, which contains references to
// interesting elements in the dialog.
function _init_widgets()
{
    if (widgets != null) {
        return; // was already called
    }
    widgets = new Object();

    widgets.pattern = document.getElementById('pattern');
    widgets.curr_pattern = widgets.pattern;
    widgets.repl_row = document.getElementById('repl-row');
    widgets.repl = document.getElementById('repl');
    widgets.curr_repl = widgets.repl;
    widgets.opt_case = document.getElementById('opt-case');

    widgets.dirs_row = document.getElementById('dirs-row');
    widgets.dirs = document.getElementById('dirs');
    widgets.subdirs_row = document.getElementById('subdirs-row');
    widgets.search_in_subdirs = document.getElementById('search-in-subdirs');
    widgets.includes_row = document.getElementById('includes-row');
    widgets.includes = document.getElementById('includes');
    widgets.excludes_row = document.getElementById('excludes-row');
    widgets.excludes = document.getElementById('excludes');

    widgets.replace_all_btn = document.getElementById('replace-all-btn');
}

let komodoInternalExtensions = {
    "Python": ["*.ksf"],
    "Perl": ["Conscript", "Construct"]
}

const _aliasLangs = {
                    "Node.js" : "JavaScript",
                    "Python3" : "Python",
                    "HTML5": "HTML"
                };
function getActualLanguageName(languageName) {
    return languageName in _aliasLangs ? _aliasLangs[languageName] : languageName;
}
            
/**
 * Initialize the dialog from `opener.ko.launch.find2_dialog_args` data.
 */
function _init() {
    var args = window.arguments[0];
    _g_searchText = args.pattern;
    widgets.pattern.value = _g_searchText;
    _g_codeintelDefn = args.defn;
    _g_view = args.view;
    _g_refactoringLanguageObj = args.refactoringLanguageObj;
    // Close this dialog when the opener goes away
    opener.addEventListener("unload", function unload(event) {
        window.close();
        event.target.removeEventListener(event.type, unload, false);
    }, false);

    // Filter function, return false if the extension is known to
    // be an internal extension used by komodo
    let keepNonKomodoInternalExtensions = function(ext) {
        var languageName;
        if (_g_codeintelDefn) {
            languageName = _g_codeintelDefn.lang;
        } else {
            try {
               languageName = _g_view.koDoc.language; 
            } catch(e) {
                log.exception(e, "failed to get language name");
            }
        }
        if (!languageName) {
            log.debug("Failed to find a languageName");
            return true;
        }
        var exts = komodoInternalExtensions[languageName];
        if (exts) {
            return exts.indexOf(ext) == -1;
        }
        return true;
    }
    // Set other dialog data (from the given args and from the
    // koIFindService.options).
    var opts = gFindSvc.options;
    widgets.repl.value = "";
    // get the language and directory, allow for a null def'n
    var languageName = null;
    var startingDir = null;
    if (_g_codeintelDefn) {
        if (_g_codeintelDefn.path) {
            startingDir = opener.ko.uriparse.dirName(_g_codeintelDefn.path);
        }
        languageName = _g_codeintelDefn.lang;
    }
    if (!startingDir) {
        try {
            startingDir = _g_view.koDoc.file.dirName;
        } catch(e) {
            log.exception(e, "Can't get dirName from current view");
            // Use the old one from the find service (yuk)
            startingDir = opts.encodedFolders;
        }
    }
    widgets.dirs.value = startingDir;
    if (!languageName) {
        try {
            languageName = _g_view.koDoc.language;
        } catch(e) {
            log.exception(e, "Can't get language from current view");
            
        }
    }
    widgets.search_in_subdirs.checked = opts.searchInSubfolders;
    {
        let extensions;
        try {
            if (!languageName) {
                throw new Error("just fill in the defaults");
            }
            var langRegistrySvc = Components.classes["@activestate.com/koLanguageRegistryService;1"].
                                   getService(Ci.koILanguageRegistryService);
            var o1 = {}, o2 = {};
            langRegistrySvc.patternsFromLanguageName(getActualLanguageName(languageName), o1, o2);
            extensions = o1.value.filter(keepNonKomodoInternalExtensions);
            widgets.includes.value = extensions.join(",");
        } catch(ex) {
            extensions = opts.encodedIncludeFiletypes.split(',');
            widgets.includes.value = opts.encodedIncludeFiletypes;
        }
        widgets.excludes.value = (opts.encodedExcludeFiletypes
                                  .split(',')
                          .filter(function(ext) extensions.indexOf(ext) === -1)
                                  .join(','));
    }
    switch (opts.caseSensitivity) {
    case koIFindOptions.FOC_INSENSITIVE:
        _set_case_widget("ignore-case");
        break;
    case koIFindOptions.FOC_SENSITIVE:
        _set_case_widget("match-case");
        break;
    case koIFindOptions.FOC_SMART:
        _set_case_widget("smart-case");
        break;
    }
    widgets.opt_case.accessKey = "c";

    // The act of opening the find dialog should reset the find session.
    // This is the behaviour of least surprise.
    var findSessionSvc = Components.classes["@activestate.com/koFindSession;1"].
                            getService(Components.interfaces.koIFindSession);
    findSessionSvc.Reset();
    // PHP Attributes: Ignore the '$' at the start of a name
    if (languageName == "PHP" && _g_searchText[0] == '$') {
        _g_searchText = _g_searchText.substring(1);
    }
    widgets.repl.select();
}

/**
 * Determine an appropriate koIFindContext instance for
 * searching/replacing, and set it to the `_g_find_context` global.
 *
 * @param {string} reason gives the reason for resetting the find context.
 *      This is only used for debugging.
 * 
 * Can return null if an appropriate context could not be determined.
 */
function reset_find_context() {
    var context = Cc["@activestate.com/koFindInFilesContext;1"]
        .createInstance(Ci.koIFindInFilesContext);
    context.type = Ci.koIFindContext.FCT_IN_FILES;

    // Use the current view's cwd for interpreting relative paths.
    if (_g_view != null &&
        _g_view.getAttribute("type") == "editor" &&
        _g_view.koDoc.file &&
        _g_view.koDoc.file.isLocal) {
        context.cwd = _g_view.koDoc.file.dirName;
    } else {
        context.cwd = gFindSvc.options.cwd;
    }
    
    _g_find_context = context;
}


function _set_case_widget(value) {
    var w = widgets.opt_case;
    switch (value) {
    case "ignore-case":
        w.value = "ignore-case";
        w.label = _bundle.GetStringFromName("matchCase");
        w.checked = false;
        w.setAttribute("tooltiptext", _bundle.GetStringFromName("ignoreCase"));
        break;
    case "match-case":
        w.value = "match-case";
        w.label = _bundle.GetStringFromName("matchCase");
        w.checked = true;
        w.setAttribute("tooltiptext", _bundle.GetStringFromName("matchCase"));
        break;
    case "smart-case":
        w.value = "smart-case";
        w.label = _bundle.GetStringFromName("smartCase");
        w.checked = true;
        w.setAttribute("tooltiptext",
            _bundle.GetStringFromName("smartCase.tooltip"));
        break;
    default:
        throw new Error("invalid case widget value: "+value);
    }
    w.accessKey = _bundle.GetStringFromName("caseCheckbox.accesskey");
}


function _toggle_collapse(widget) {
    if (widget.hasAttribute("collapsed")) {
        widget.removeAttribute("collapsed");
    } else {
        widget.setAttribute("collapsed", "true");
    }
}

function _collapse_widget(widget, collapse) {
    if (collapse) {
        widget.setAttribute("collapsed", "true");
    } else {
        if (widget.hasAttribute("collapsed"))
            widget.removeAttribute("collapsed");
    }
}

function _hide_widget(widget, hide) {
    if (hide) {
        widget.setAttribute("hidden", "true");
    } else {
        if (widget.hasAttribute("hidden"))
            widget.removeAttribute("hidden");
    }
}

function _disable_widget(widget) {
    widget.setAttribute("disabled", "true");
}
function _enable_widget(widget) {
    if (widget.hasAttribute("disabled")) {
        widget.removeAttribute("disabled");
    }
}

/**
 * Toggle whether the window is raised
 */
function toggle_pin() {
    var pinned = widgets.pin_btn.checked;
    _g_prefs.setBooleanPref("find-pinFindReplaceDialog", pinned);
    pinDialog(pinned);
}

function pinDialog(pinned) {
    function getXULWindowForDOMWindow(win)
        win.QueryInterface(Ci.nsIInterfaceRequestor)
           .getInterface(Ci.nsIWebNavigation)
           .QueryInterface(Ci.nsIDocShellTreeItem)
           .treeOwner
           .QueryInterface(Ci.nsIInterfaceRequestor)
           .getInterface(Ci.nsIXULWindow);
    let rootWin = getXULWindowForDOMWindow(window);
    let parentWin = ((opener && !opener.closed)
                     ? getXULWindowForDOMWindow(opener)
                     : null);
    try {
        Cc["@activestate.com/koIWindowManagerUtils;1"]
          .getService(Ci.koIWindowManagerUtils)
          .setOnTop(rootWin, parentWin, pinned);
    } catch(ex) {
        log.exception(ex, "pinDialog: Can't setOnTop");
    }
}


//---- Rx-integration stuff

// #if WITH_RX
function open_in_rx()
{
    try {
        // Get the appropriate Rx data from the current find settings and
        // the current view, if any.
        // search_text: if have a current editor view, then use
        var mode = null;
        var replacement = null;
        var regex = widgets.pattern.value;
        mode = "replace-all";
        replacement = widgets.repl.value;
        var match_case = !widgets.opt_case.checked;
        // Search text:
        // - if no editor view, no search_text
        // - else if "Search In: current selection" then get the selection,
        // - else if file < LINE_LIMIT lines, use whole file,
        // - else use LINE_LIMIT/2 lines either side of the current file.
        var search_text = null;
        var scimoz = null;
        const LINE_LIMIT = 1000;
        try {
            scimoz = _g_view.scintilla.scimoz;
        } catch(ex) {
            /* pass: just don't have a current editor view */
        }
        if (scimoz) {
            search_text = scimoz.selText;
            if (!search_text) {  /* no selection */
                if (scimoz.lineCount < LINE_LIMIT) {
                    search_text = scimoz.text;
                } else {
                    var currLine = scimoz.lineFromPosition(scimoz.currentPos);
                    var startLine = Math.max(0, currLine - (LINE_LIMIT/2));
                    var endLine = Math.min(scimoz.lineCount,
                                           currLine + (LINE_LIMIT/2));
                    var startPos = scimoz.charPosAtPosition(
                        scimoz.positionFromLine(startLine));
                    var endPos = scimoz.charPosAtPosition(
                        scimoz.positionFromLine(endLine));
                    if (startLine) {
                        search_text = "(" + startLine + " lines skipped)\n";
                    } else {
                        search_text = ""
                    }
                    search_text += scimoz.getTextRange(startPos, endPos);
                    if (endLine < scimoz.lineCount) {
                        if (search_text[search_text.length-1] != "\n") {
                            search_text += "\n";
                        }
                        search_text += "(" + (scimoz.lineCount-endLine)
                                      + " lines skipped)\n";
                    }
                }
            }
        }

        // Open Rx.
        var retval = ko.launch.rxToolkit(
            mode, regex, replacement, search_text, match_case,
            true /* isModal */,
            true /* flagMultiline */);
        
        // Update find dialog from changes in Rx.
        var scimoz_eol_mode = (scimoz ? scimoz.eOLMode : null);
        var escaped_text;
        if (retval.regex != regex) {
            escaped_text = _escape_for_textbox(retval.regex, scimoz_eol_mode);
            // Update both single and multiline replacement fields.
            widgets.pattern.value = escaped_text;
        }
        if (retval.replacement != replacement)
        {
            escaped_text = _escape_for_textbox(retval.replacement, scimoz_eol_mode);
            // Update both single and multiline replacement fields.
            widgets.repl.value = escaped_text;
        }
        widgets.pattern.focus();
    } catch (ex) {
        log.exception(ex);
    }
}

/**
 * Escape the given string for putting in a single-line textbox.
 * Primarily we put in C-escapes for EOL chars. We also try to normalize
 * EOLs to the EOL-style of the current view, if any. We do this because
 * Rx (from which these strings are coming) standardizes on '\n' EOLs -- even
 * on Windows.
 */
function _escape_for_textbox(s, scimoz_eol_mode /* =null */)
{
    if (typeof(scimoz_eol_mode) == "undefined") scimoz_eol_mode = null;
    if (scimoz_eol_mode == Components.interfaces.ISciMoz.SC_EOL_CRLF) {
        s = s.replace('\r\n', '\n').replace('\n', '\r\n');
    }
    return s.replace('\n', '\\n').replace('\r', '\\r');
}
// #endif
